@@ -30,7 +30,7 @@
 //!     let mut client = RestClient::new("http://httpbin.org").unwrap();
 //!
 //!     // GET http://httpbin.org/anything and deserialize the result automatically
-//!     let data: HttpBinAnything = client.get(()).await.unwrap();
+//!     let data = client.get::<_, HttpBinAnything>(()).await.unwrap().into_inner();
 //!     println!("{:?}", data);
 //! }
 //! ```
@@ -42,6 +42,7 @@ use hyper::{Client, Method, Request};
 use hyper_tls::HttpsConnector;
 use log::{debug, trace, error};
 use std::{error, fmt};
+use std::ops::Deref;
 use std::time::Duration;
 use url::Url;
 
@@ -65,13 +66,60 @@ pub type Query<'a> = [(&'a str, &'a str)];
 
 pub type HyperClient = Client<HttpsConnector<hyper::client::HttpConnector>>;
 
+/// Type returned by client query functions
+#[derive(Debug)]
+pub struct Response<T> {
+    body: T,
+    headers: HeaderMap,
+}
+
+impl<T> Response<T> {
+    /// Unwraps the response, getting the owned inner body
+    pub fn into_inner(self) -> T {
+        self.body
+    }
+
+    /// Response headers sent by the server
+    pub fn headers(&self) -> &HeaderMap {
+        &self.headers
+    }
+}
+
+impl Response<String> {
+    /// Parse a response body
+    fn parse<T: serde::de::DeserializeOwned>(self) -> Result<Response<T>, Error> {
+        #[cfg(feature = "lib-serde-json")]
+        {
+            let Self { body, headers } = self;
+            serde_json::from_str(&body)
+                .map(|body| Response { body, headers })
+                .map_err(|err| Error::DeserializeParseError(err, body))
+        }
+
+        #[cfg(feature = "lib-simd-json")]
+        {
+            let Self { mut body, headers } = self;
+            simd_json::serde::from_str(&mut body)
+                .map(|body| Response { body, headers })
+                .map_err(|err| Error::DeserializeParseSimdJsonError(err, body))
+        }
+    }
+}
+
+impl<T> Deref for Response<T> {
+    type Target = T;
+
+    fn deref(&self) -> &Self::Target {
+        &self.body
+    }
+}
+
 /// REST client to make HTTP GET and POST requests.
 pub struct RestClient {
     client: HyperClient,
     baseurl: url::Url,
     auth: Option<String>,
     headers: HeaderMap,
-    response_headers: HeaderMap,
     timeout: Duration,
     send_null_body: bool,
     body_wash_fn: fn(String) -> String,
@@ -269,7 +317,6 @@ impl RestClient {
             baseurl,
             auth: None,
             headers: HeaderMap::new(),
-            response_headers: HeaderMap::new(),
             timeout: builder.timeout,
             send_null_body: builder.send_null_body,
             body_wash_fn: std::convert::identity,
@@ -320,53 +367,30 @@ impl RestClient {
         self.headers.clear();
     }
 
-    /// Response headers captured from previous request
-    pub fn response_headers(&mut self) -> &HeaderMap {
-        &self.response_headers
-    }
-
-    /// Make a GET request
-    pub async fn get<U, T>(&mut self, params: U) -> Result<T, Error>
+    /// Make a GET request.
+    pub async fn get<U, T>(&self, params: U) -> Result<Response<T>, Error>
     where
         T: serde::de::DeserializeOwned + RestPath<U>,
     {
         let req = self.make_request::<U, T>(Method::GET, params, None, None)?;
+        let res = self.run_request(req).await?;
 
-        #[cfg(feature = "lib-serde-json")]
-        {
-            let body = self.run_request(req).await?;
-            serde_json::from_str(body.as_str()).map_err(|err| Error::DeserializeParseError(err, body))
-        }
-
-        #[cfg(feature = "lib-simd-json")]
-        {
-            let mut body = self.run_request(req).await?;
-            simd_json::serde::from_str(&mut body).map_err(|err| Error::DeserializeParseSimdJsonError(err, body))
-        }
+        res.parse()
     }
 
     /// Make a GET request with query parameters.
-    pub async fn get_with<U, T>(&mut self, params: U, query: &Query<'_>) -> Result<T, Error>
+    pub async fn get_with<U, T>(&self, params: U, query: &Query<'_>) -> Result<Response<T>, Error>
     where
         T: serde::de::DeserializeOwned + RestPath<U>,
     {
         let req = self.make_request::<U, T>(Method::GET, params, Some(query), None)?;
-        
-        #[cfg(feature = "lib-serde-json")]
-        {
-            let body = self.run_request(req).await?;
-            serde_json::from_str(body.as_str()).map_err(|err| Error::DeserializeParseError(err, body))
-        }
-        
-        #[cfg(feature = "lib-simd-json")]
-        {
-            let mut body = self.run_request(req).await?;
-            simd_json::serde::from_str(&mut body).map_err(|err| Error::DeserializeParseSimdJsonError(err, body))
-        }
+        let res = self.run_request(req).await?;
+
+        res.parse()
     }
 
     /// Make a POST request.
-    pub async fn post<U, T>(&mut self, params: U, data: &T) -> Result<(), Error>
+    pub async fn post<U, T>(&self, params: U, data: &T) -> Result<Response<()>, Error>
     where
         T: serde::Serialize + RestPath<U>,
     {
@@ -374,7 +398,7 @@ impl RestClient {
     }
 
     /// Make a PUT request.
-    pub async fn put<U, T>(&mut self, params: U, data: &T) -> Result<(), Error>
+    pub async fn put<U, T>(&self, params: U, data: &T) -> Result<Response<()>, Error>
     where
         T: serde::Serialize + RestPath<U>,
     {
@@ -382,26 +406,26 @@ impl RestClient {
     }
 
     /// Make a PATCH request.
-    pub async fn patch<U, T>(&mut self, params: U, data: &T) -> Result<(), Error>
+    pub async fn patch<U, T>(&self, params: U, data: &T) -> Result<Response<()>, Error>
     where
         T: serde::Serialize + RestPath<U>,
     {
         self.post_or_put(Method::PATCH, params, data).await
     }
 
-    async fn post_or_put<U, T>(&mut self, method: Method, params: U, data: &T) -> Result<(), Error>
+    async fn post_or_put<U, T>(&self, method: Method, params: U, data: &T) -> Result<Response<()>, Error>
     where
         T: serde::Serialize + RestPath<U>,
     {
         let data = serde_json::to_string(data).map_err(Error::SerializeParseError)?;
 
         let req = self.make_request::<U, T>(method, params, None, Some(data))?;
-        self.run_request(req).await?;
-        Ok(())
+        let res = self.run_request(req).await?;
+        Ok(Response { body: (), headers: res.headers })
     }
 
     /// Make POST request with query parameters.
-    pub async fn post_with<U, T>(&mut self, params: U, data: &T, query: &Query<'_>) -> Result<(), Error>
+    pub async fn post_with<U, T>(&self, params: U, data: &T, query: &Query<'_>) -> Result<Response<()>, Error>
     where
         T: serde::Serialize + RestPath<U>,
     {
@@ -409,7 +433,7 @@ impl RestClient {
     }
 
     /// Make PUT request with query parameters.
-    pub async fn put_with<U, T>(&mut self, params: U, data: &T, query: &Query<'_>) -> Result<(), Error>
+    pub async fn put_with<U, T>(&self, params: U, data: &T, query: &Query<'_>) -> Result<Response<()>, Error>
     where
         T: serde::Serialize + RestPath<U>,
     {
@@ -417,7 +441,7 @@ impl RestClient {
     }
 
     /// Make PATCH request with query parameters.
-    pub async fn patch_with<U, T>(&mut self, params: U, data: &T, query: &Query<'_>) -> Result<(), Error>
+    pub async fn patch_with<U, T>(&self, params: U, data: &T, query: &Query<'_>) -> Result<Response<()>, Error>
     where
         T: serde::Serialize + RestPath<U>,
     {
@@ -425,24 +449,24 @@ impl RestClient {
     }
 
     async fn post_or_put_with<U, T>(
-        &mut self,
+        &self,
         method: Method,
         params: U,
         data: &T,
         query: &Query<'_>,
-    ) -> Result<(), Error>
+    ) -> Result<Response<()>, Error>
     where
         T: serde::Serialize + RestPath<U>,
     {
         let data = serde_json::to_string(data).map_err(Error::SerializeParseError)?;
 
         let req = self.make_request::<U, T>(method, params, Some(query), Some(data))?;
-        self.run_request(req).await?;
-        Ok(())
+        let res = self.run_request(req).await?;
+        Ok(Response { body: (), headers: res.headers })
     }
 
     /// Make a POST request and capture returned body.
-    pub async fn post_capture<U, T, K>(&mut self, params: U, data: &T) -> Result<K, Error>
+    pub async fn post_capture<U, T, K>(&self, params: U, data: &T) -> Result<Response<K>, Error>
     where
         T: serde::Serialize + RestPath<U>,
         K: serde::de::DeserializeOwned,
@@ -451,7 +475,7 @@ impl RestClient {
     }
 
     /// Make a PUT request and capture returned body.
-    pub async fn put_capture<U, T, K>(&mut self, params: U, data: &T) -> Result<K, Error>
+    pub async fn put_capture<U, T, K>(&self, params: U, data: &T) -> Result<Response<K>, Error>
     where
         T: serde::Serialize + RestPath<U>,
         K: serde::de::DeserializeOwned,
@@ -460,11 +484,11 @@ impl RestClient {
     }
 
     async fn post_or_put_capture<U, T, K>(
-        &mut self,
+        &self,
         method: Method,
         params: U,
         data: &T,
-    ) -> Result<K, Error>
+    ) -> Result<Response<K>, Error>
     where
         T: serde::Serialize + RestPath<U>,
         K: serde::de::DeserializeOwned,
@@ -472,17 +496,17 @@ impl RestClient {
         let data = serde_json::to_string(data).map_err(Error::SerializeParseError)?;
 
         let req = self.make_request::<U, T>(method, params, None, Some(data))?;
-        let body = self.run_request(req).await?;
-        serde_json::from_str(body.as_str()).map_err(|err| Error::DeserializeParseError(err, body))
+        let res = self.run_request(req).await?;
+        res.parse()
     }
 
     /// Make a POST request with query parameters and capture returned body.
     pub async fn post_capture_with<U, T, K>(
-        &mut self,
+        &self,
         params: U,
         data: &T,
         query: &Query<'_>,
-    ) -> Result<K, Error>
+    ) -> Result<Response<K>, Error>
     where
         T: serde::Serialize + RestPath<U>,
         K: serde::de::DeserializeOwned,
@@ -492,11 +516,11 @@ impl RestClient {
 
     /// Make a PUT request with query parameters and capture returned body.
     pub async fn put_capture_with<U, T, K>(
-        &mut self,
+        &self,
         params: U,
         data: &T,
         query: &Query<'_>,
-    ) -> Result<K, Error>
+    ) -> Result<Response<K>, Error>
     where
         T: serde::Serialize + RestPath<U>,
         K: serde::de::DeserializeOwned,
@@ -505,12 +529,12 @@ impl RestClient {
     }
 
     async fn post_or_put_capture_with<U, T, K>(
-        &mut self,
+        &self,
         method: Method,
         params: U,
         data: &T,
         query: &Query<'_>,
-    ) -> Result<K, Error>
+    ) -> Result<Response<K>, Error>
     where
         T: serde::Serialize + RestPath<U>,
         K: serde::de::DeserializeOwned,
@@ -518,32 +542,32 @@ impl RestClient {
         let data = serde_json::to_string(data).map_err(Error::SerializeParseError)?;
 
         let req = self.make_request::<U, T>(method, params, Some(query), Some(data))?;
-        let body = self.run_request(req).await?;
-        serde_json::from_str(body.as_str()).map_err(|err| Error::DeserializeParseError(err, body))
+        let res = self.run_request(req).await?;
+        res.parse()
     }
 
     /// Make a DELETE request.
-    pub async fn delete<U, T>(&mut self, params: U) -> Result<(), Error>
+    pub async fn delete<U, T>(&self, params: U) -> Result<Response<()>, Error>
     where
         T: RestPath<U>,
     {
         let req = self.make_request::<U, T>(Method::DELETE, params, None, None)?;
-        self.run_request(req).await?;
-        Ok(())
+        let res = self.run_request(req).await?;
+        Ok(Response { body: (), headers: res.headers })
     }
 
     /// Make a DELETE request with query and body.
-    pub async fn delete_with<U, T>(&mut self, params: U, data: &T, query: &Query<'_>) -> Result<(), Error>
+    pub async fn delete_with<U, T>(&self, params: U, data: &T, query: &Query<'_>) -> Result<Response<()>, Error>
     where
         T: serde::Serialize + RestPath<U>,
     {
         let data = serde_json::to_string(data).map_err(Error::SerializeParseError)?;
         let req = self.make_request::<U, T>(Method::DELETE, params, Some(query), Some(data))?;
-        self.run_request(req).await?;
-        Ok(())
+        let res = self.run_request(req).await?;
+        Ok(Response { body: (), headers: res.headers })
     }
 
-    async fn run_request(&mut self, req: hyper::Request<hyper::Body>) -> Result<String, Error> {
+    async fn run_request(&self, req: hyper::Request<hyper::Body>) -> Result<Response<String>, Error> {
         debug!("{} {}", req.method(), req.uri());
         trace!("{:?}", req);
 
@@ -551,14 +575,14 @@ impl RestClient {
         let work = async {
             let res = self.client.request(req).await?;
 
-            self.response_headers = res.headers().clone();
+            let response_headers = res.headers().clone();
             let status = res.status();
             let mut body = hyper::body::aggregate(res).await?;
             let body = body.copy_to_bytes(body.remaining());
 
             let body = String::from_utf8_lossy(&body);
 
-            Ok::<_, hyper::Error>((body.to_string(), status))
+            Ok::<_, hyper::Error>((response_headers, body.to_string(), status))
         };
 
         let res;
@@ -568,20 +592,20 @@ impl RestClient {
             res = work.await?;
         }
 
-        let (body, status) = res;
+        let (response_headers, body, status) = res;
 
         if !status.is_success() {
             error!("server returned \"{}\" error", status);
             return Err(Error::HttpError(status.as_u16(), body));
         }
 
-        trace!("response headers: {:?}", self.response_headers);
+        trace!("response headers: {:?}", response_headers);
         trace!("response body: {}", body);
-        Ok((self.body_wash_fn)(body))
+        Ok(Response { body: (self.body_wash_fn)(body), headers: response_headers })
     }
 
     fn make_request<U, T>(
-        &mut self,
+        &self,
         method: Method,
         params: U,
         query: Option<&Query>,
